#######################################################
# R implementation of W-RF for 
# "Tree-based Ensemble Methods For Individualized Treatment Rules" 
# Authors: Kehao Zhu, Ying Huang, Xiao-Hua (Andrew) Zhou
# For questions, please contact Kehao Zhu, zhukehao@uw.edu
#
#######################################################
library(compiler);library(Rcpp) #load required packages.

w.rf<-function(dat,ntree=100,x.try,min.node.size){
  # Function to implement W-RF
  # Args: 
  #   dat: a matrix of data consisting Xs (covariates), 
  #        y (estimate class label) and wt (weight)
  #   x.try:  (aka mtry) number of Xs considered to split
  #   min.node.size: minimal node size
  # Returns: a list of 3 elements:
  # 1. Matrix represent tree-based models
  # 2. OOB prediction without / with X_j permuted
  # 3. Type 1 Variable importance
  
  
  #repetitvely call c.one.forest to build a single boostrapped tree
  results<-replicate(ntree,
           c.one.forest(dat=dat,x.try=x.try,min.node.size=min.node.size))
  
  oob.pred<-apply(simplify2array(results[2,]),
                  c(1,2), FUN=function(x)(mean(x,na.rm=T)>0.5)) 
  
  impurity.reduce<-apply(simplify2array(results[3,]),
                         c(1,2),FUN=function(x)(mean(x,na.rm=T)))

  trees<-results[1,]
  return(list(trees,oob.pred,impurity.reduce))
}

c.w.rf<-cmpfun(w.rf) #complied the function for better performance



library(glmnet);library(randomForest)
data.prepare<-function(dat,type,ntree)
{
  # Function to estimate individual weights and
  # labels for bootstrapped data
  # Args: 
  #   dat: a dataframe consisting Xs (covariates), 
  # 	   Y (outcome) and A (trt assignment)
  #   type: type of weights
  #   min.node.size: minimal node size
  # Returns: a matrix of data consisting Xs (covariates), 
  #        y* (estimate class label) and wt (weight)
  dat<-as.data.frame(dat)
  p<-ncol(dat)-2
  if(type=="IPWE"){
    wt<- 2*(dat$y*dat$t-dat$y*(1-dat$t))
  }
  if(type=="AIPWE.lm")
  {
    y<-dat[,"y"]
    xx<-dat[,1:p]
    xx.f<-cbind(1,xx,xx*dat[,"t"],t=dat[,"t"]) 
    fit<-glm.fit(x=xx.f,y=y)
    xx1<-cbind(1,xx,xx*1,t=1);xx0<-cbind(1,xx,xx*0,t=0)
    fit.coeff<-t(as.matrix(coef(fit)))
    u1<-fit.coeff %*% t(xx1)
    u0<-fit.coeff %*% t(xx0)
    wt<- as.numeric (2*(dat$y*dat$t-dat$y*(1-dat$t)-
		      (dat$t-0.5)*u1-(dat$t-0.5)*u0))
  }

  if(type=='AIPWE.rf')
  {
    u<-rf.mu(dat=dat,ntree=ntree)
    wt<- 2*(dat$y*dat$t-dat$y*(1-dat$t)-(dat$t-0.5)*u[,1]-(dat$t-0.5)*u[,2])
  }
  
  dat.out<-as.data.frame(dat[,c(1:p)])
  dat.out$wt<-abs(wt)
  dat.out$y<-as.numeric(wt>0)
  return(as.matrix(dat.out))
}

rf.mu<-function(dat,ntree)
{
  dat<-as.data.frame(dat)
  p<-ncol(dat)-1
  mtry.b=round(2/3 * p)

  model.rf<-randomForest(y~.,mtry=mtry.b,ntree=ntree,data=dat)
  data.rf0<-dat;data.rf0$t<-0
  data.rf1<-dat;data.rf1$t<-1
  u1<-predict(model.rf,newdata=data.rf1)
  u0<-predict(model.rf,newdata=data.rf0)
  return(cbind(u1,u0))
}


one.forest<-function(dat,x.try,min.node.size)
{
  # Function to bootstrap samples and build a single tree
  # Args: 
  #   dat: a matrix of data consisting Xs (covariates), 
  #        y (estimate class label) and wt (weight)
  #   x.try:  (aka mtry) number of Xs considered to split
  #   min.node.size: minimal node size
  # Returns: a list of elements:
  # 1. Matrix represent a single tree-based model
  # 2. OOB prediction without / with X_j permuted
  # 3. Type 1 Variable importance
  
  N<-nrow(dat)
  p<-ncol(dat)-2 #number of predictors
  dat.ind<-seq(1:N)
  # sample index with replacement (boostrap)
  boot.ind<-sample(dat.ind,N,replace=T)
  # out-of-bag (OOB) index 
  oob.ind<-setdiff(dat.ind,boot.ind)

  # estimate individual weight and label
  datBoot<-data.prepare(dat=dat[boot.ind,],type='AIPWE.rf',ntree=500)

  # call c.mytree to build a single tree
  tmp<-c.mytree(dat=datBoot,
                min.node.size=min.node.size,x.try=x.try)
  one.tree<-tmp[[1]]
  oob.dat<-dat[oob.ind,]
  
  # OOB predictions with and without X_j permuted
  oob.prediction<-matrix(nrow=N,ncol=p+1)
  oob.prediction[oob.ind,1]<-pred_tree_cpp(as.matrix(one.tree),oob.dat)
  for(i in 1:p)
  {
    permute.oob<-oob.dat
    permute.oob[,i]<-sample(oob.dat[,i],replace=F)
    oob.prediction[oob.ind,i+1]<- pred_tree_cpp(one.tree,permute.oob)
  }
  
  return(list(one.tree,oob.prediction,tmp[[2]]))
}

c.one.forest<-cmpfun(one.forest)

mytree<-function(dat,min.node.size,x.try)
{
  # Function to build a single tree
  # Args: 
  #   dat: a matrix of data consisting Xs (covariates), 
  #        y (estimate class label) and wt (weight)
  #   x.try:  (aka mtry) number of Xs considered to split
  #   min.node.size: minimal node size
  # Returns: a list of elements:
  # 1. Matrix represent a single tree-based model
  # 2. OOB prediction without / with X_j permuted
  # 3. Type 1 Variable importance
  
  #maximal node size
  node<-seq(1:(2*nrow(dat)-min.node.size+2)) 
  
  #predictor index selected to be splitted
  x.index<-seq(1:(ncol(dat)-2))  
  
  # a matrix representing a tree model 
  tree.matrix<-cbind(node,feature=NA,
                     cutoff=NA,impurity=NA,
                     children.impurity=NA,
                     left.node=NA,
                     right.node=NA,a=NA,terminal=-1,Stop=0)
  # list of data in each tree node
  list.dat<-list(dat)
  
  #initiate the loop
  tree.matrix[1,"terminal"]<-1
  tree.matrix[1,"impurity"]<-impurity.dat(list.dat[[1]])
  new.node<-1
  
  #loop to implement recursive binary splitting algorithm  
  while(sum(tree.matrix[,"terminal"]==1 & tree.matrix[,"Stop"]==0)!=0)
  {
    terminal.node<-tree.matrix[tree.matrix[,"terminal"]==1 & 
                                 tree.matrix[,"Stop"]==0,"node"]
    split.matrix<-cbind(node,x.ind=NA,cutoff=NA,children.impurity=NA)
    
    for(j in terminal.node)
    {
      if(nrow(list.dat[[j]])>min.node.size & 
         impurity.dat(dat=list.dat[[j]])>0) 
      {
        split.matrix[j,c("x.ind","cutoff","children.impurity")]<-
          c.compare(x.ind=sort(sample(x.index,size=x.try,replace=F)),
                    mydat=list.dat[[j]])
        
      } else 
      {
        
        tree.matrix[j,"Stop"]<-1
        terminal.data<-list.dat[[j]]
        tree.matrix[j,c("a")] <- ifelse(
          sum((terminal.data[,"y"]-1)^2*terminal.data[,"wt"])<
          sum((terminal.data[,"y"]-0)^2*terminal.data[,"wt"]),1,0)
        
      }
    }
    
    
    if (prod(tree.matrix[tree.matrix[,"terminal"]==1,"Stop"])){break}
    #the largest decrease
    impurity.descrease<-tree.matrix[,"impurity"]-
      split.matrix[,"children.impurity"]
    split.index.multi<-as.vector(which(
      impurity.descrease==max(impurity.descrease,na.rm=T)))
    split.index<-ifelse(length(split.index.multi)>1,
                        sample(x=split.index.multi,size=1),split.index.multi)
    split.node<-split.matrix[split.index,]
    
    
    tree.matrix[tree.matrix[,"node"]==split.node["node"],
                c("feature","cutoff","children.impurity","terminal")]<-
      c(split.node[-1],1)
    
    
    split.data<-list.dat[[split.node["node"]]]
    list.dat[[new.node+1]]<-split.data[split.data[,split.node["x.ind"]]<=
                                         split.node["cutoff"],] 
    list.dat[[new.node+2]]<-split.data[split.data[,split.node["x.ind"]]>
                                         split.node["cutoff"],]
    if(class(list.dat[[new.node+1]])=='numeric')
    {
      list.dat[[new.node+1]]<-t(as.matrix(list.dat[[new.node+1]]))
    }
    if(class(list.dat[[new.node+2]])=='numeric')
    {
      list.dat[[new.node+2]]<-t(as.matrix(list.dat[[new.node+2]]))
    }
    
    
    tree.matrix[tree.matrix[,"node"]==split.node["node"],
                c("left.node","right.node","terminal")]<-
                c(new.node+1,new.node+2,0)
    tree.matrix[c(new.node+1,new.node+2),"terminal"]<-c(1,1)
    tree.matrix[new.node+1,"impurity"]<-impurity.dat(list.dat[[new.node+1]])
    tree.matrix[new.node+2,"impurity"]<-impurity.dat(list.dat[[new.node+2]])
    new.node<-new.node+2
    
  }
  
  
  #Type 1 VI
  tmp<-tree.matrix[tree.matrix[,"terminal"]!=-1,]
  tmp<-tmp[is.na(tmp[,"feature"])==F,
           c("feature","impurity","children.impurity")]
  if(class(tmp)=="matrix")
  {
    reduced<-tmp[,"impurity"]-tmp[,"children.impurity"]
    red.im<-as.matrix(xtabs(reduced~tmp[,"feature"]))
  } else if (class(tmp)=="numeric")
  {
    reduced<-tmp["impurity"]-tmp["children.impurity"]
    red.im<-as.matrix(xtabs(reduced~tmp["feature"]))
  }
  
  vi<-rbind(x.index,0)
  for(i in x.index)
  {
    if(sum(rownames(red.im)==i)!=0)
    {
      vi[2,i]<-red.im[rownames(red.im)==i]
    } else next
  }
  
  
  #formating the final tree model in the matrix format
  tree.matrix<-tree.matrix[,c("node","feature","cutoff",
                            "left.node","right.node","a","terminal")]
  tree.matrix<-tree.matrix[tree.matrix[,"terminal"]!=-1,]
  tree.matrix[is.na(tree.matrix)]<- -1
  if(class(tree.matrix)=="numeric"){tree.matrix = t(as.matrix(tree.matrix))}
  
  return(list(tree.matrix,vi))
}

c.mytree<-cmpfun(mytree)

impurity.dat<-function(dat)
{
  # Function to calculate impurity / weighted 
  # misclassifcate errors of a node
  # Args: 
  #   dat: a matrix of data consisting Xs (covariates), 
  #        y (estimate class label) and wt (weight)
  # Returns: 
  #   weighted misclassifcate errors
  
  wt<-dat[,"wt"]
  y<-dat[,"y"]
  min(sum(wt*(y-1)^2),sum(wt*(y-0)^2))
}

scan<-function(mydat,feature){
  
  # Function to scan a covariate and find the best cutoff point
  # Args: 
  #   mydat: data in the node
  #   feature: the covariate considered to split
  # Returns: 
  #   The best cutoff point and corresponding impurity measure
  
  mydat<-mydat[order(mydat[,feature]),]
  N<-nrow(mydat)
  impurity.l<-numeric(N-1)
  impurity.r<-numeric(N-1)
  sequence<-seq(1:(N-1))
  yy<-mydat[,"y"]
  wt<-mydat[,"wt"]
  
  tmp<-sapply(sequence,c.for.scan,yy=yy,wt=wt,N=N)
  
  impurity.l<-tmp[1,]
  impurity.r<-tmp[2,]
  index.multi<-which((impurity.l+impurity.r) == min(impurity.l+impurity.r))
  index<-ifelse(length(index.multi)>1,sample(index.multi,size=1),index.multi)
  l.cutoff<-mydat[index, feature]
  c.impurity<-min(impurity.l+impurity.r)
  return(c(l.cutoff,c.impurity))
}

c.scan<-cmpfun(scan)

for.scan<-function(i,yy,wt,N){
  
  # Function to calculate the impurity of particular split point
  # Args: 
  #   i: index to split
  #   yy: class label
  #   wt: weight
  #   N: total number of observation in this node
  # Returns: 
  #   impurity of particular split point
  
  y.l<-yy[1:i]
  wt.l<-wt[1:i]
  y.r<-yy[(i+1):N]
  wt.r<-wt[(i+1):N]
  l.imp<-min(sum(wt.l*(y.l-1)^2),sum(wt.l*(y.l-0)^2))
  r.imp<-min(sum(wt.r*(y.r-1)^2),sum(wt.r*(y.r-0)^2))
  return(c(l.imp,r.imp))
}

c.for.scan<-cmpfun(for.scan)

compare<-function(x.ind,mydat)
{  
  # Function to find the best covariate and cutoff point combination
  # Args: 
  #   x.ind: covariates randomly selected to be the candidate
  #   mydat: data in the node
  # Returns: 
  #   the best covariate and cutoff point
  
  all.cut<-matrix(ncol=2,nrow=length(x.ind))
  for (index in 1:length(x.ind))
  {
    all.cut[index,c(1:2)]<-c.scan(mydat=mydat,feature=x.ind[index])
  }
  
  index.many<-which(all.cut[,2]==min(all.cut[,2],na.rm=T))
  #randomly select one node if equal
  index<-ifelse(length(index.many)>1,sample(x=index.many,size=1),index.many) 
  s.feature<-x.ind[index]
  return(c(s.feature,all.cut[index,]))
}

c.compare<-cmpfun(compare)

cppFunction('NumericVector pred_tree_cpp(NumericMatrix tree,
            NumericMatrix data){
            // Function to predict the classification 
            // label from a single tree
            // Args: 
            //   tree: a matrix representing a tree
            //   data: a matrix of (testing) data
            // Returns:
            //   yhat: the predicted class label 

            int obs = data.nrow();
            NumericVector out(obs);
            int node=0;
            
            for(int i = 0; i< obs; ++i){
            node = 0;
            while(TRUE){
              if(tree(node,6) == 1){
                 out[i] = tree(node,5);
                 break;
                }
              else if ( data(i,(tree(node,1)-1)) > tree(node,2) ) 
                {
                  node = (tree(node,4)-1);
                } else 
                {
                  node = (tree(node,3)-1);
                }
              }
            }
            return out;
            }')


pred.forest<-function(dat,forest.obj)
{
  # Function to predict the classification label from a forest
  # Args: 
  #   dat: a matrix of data consisting of covariates / predictors
  #   forest.obj: a list return from w.rf
  # Returns: 
  #    the predicted class label / treatment rule
  
  ntree<-length(forest.obj[[1]])
  as.numeric(t(rowMeans(sapply(seq(1:ntree),
  FUN=function(x){pred_tree_cpp(as.matrix(forest.obj[[1]][[x]]),dat)})
  )>0.5))
}

c.pred.forest<-cmpfun(pred.forest)